#ifndef _TABULAR_OUTPUT_H
#define  _TABULAR_OUTPUT_H

#define UTIL_T_MAX_CELLS 20
#define UTIL_T_MAX_CELLCONTENT_LEN 30
#define UTIL_T_NUMALLOC_ROW 10

#define UTIL_T_CONTENTTYPE_STRING 0
#define UTIL_T_CONTENTTYPE_NUMERIC 1

#define UTIL_T_TYPE_ASCII 0
#define UTIL_T_TYPE_CSV 1
#define UTIL_T_TYPE_HTML 2
#define UTIL_T_TYPE_MAX 2

#define UTIL_T_CELLSEPARATOR '|'
#define UTIL_T_ROWSEPARATOR '-'

#define UTIL_T_TYPE_HELP "output-format of table: 0: ASCII, 1: CSV, 2 : HTML"
#define UTIL_T_SORT_HELP "table-column to sort"

#endif /* _TABULAR_OUTPUT_H */
