/*
 * Copyright 2000, International Business Machines Corporation and others.
 * All Rights Reserved.
 *
 * This software has been released under the terms of the IBM Public
 * License.  For details, see the LICENSE file in the top-level source
 * directory or online at http://www.openafs.org/dl/license10.html
 */

#ifndef _AFS_CONSTS_H_
#define _AFS_CONSTS_H_

#define	AFS_MAXHOSTS		13	/* max hosts per single volume */
#define	AFS_OMAXHOSTS	 	8	/* backwards compatibility */
#define AFS_MAXCELLHOSTS	8	/* max vldb servers per cell */

#define AFS_PIOCTL_MAXSIZE		2048	/* max returned from PIOCTL */

#endif /* _AFS_CONSTS_H_ */
