EXPORTS
	hc_DES_cbc_encrypt                      @1
	hc_DES_cbc_cksum                        @2
	hc_DES_ecb_encrypt                      @3
	hc_DES_init_random_number_generator     @4
	hc_DES_key_sched                        @5
	hc_DES_new_random_key                   @6
	hc_DES_pcbc_encrypt                     @7
	hc_DES_set_odd_parity                   @8
	hc_DES_string_to_key                    @9
	hc_MD4_Final                            @10
	hc_MD4_Init                             @11
	hc_MD4_Update                           @12
	hc_MD5_Final                            @13
	hc_MD5_Init                             @14
	hc_MD5_Update                           @15
	hc_UI_UTIL_read_pw_string               @16
	hc_DES_is_weak_key			@17
	hc_DES_check_key_parity			@18
	hc_EVP_CIPHER_CTX_block_size		@19
	hc_EVP_CIPHER_CTX_cleanup		@20
	hc_EVP_CIPHER_CTX_init			@21
	hc_EVP_CIPHER_CTX_iv_length		@22
	hc_EVP_Cipher				@23
	hc_EVP_CipherInit_ex			@24
	hc_EVP_Digest				@25
	hc_EVP_DigestFinal_ex			@26
	hc_EVP_DigestInit_ex			@27
	hc_EVP_DigestUpdate			@28
	hc_EVP_MD_CTX_create			@29
	hc_EVP_MD_CTX_destroy			@30
	hc_EVP_aes_128_cbc			@31
	hc_EVP_aes_256_cbc			@32
	hc_EVP_des_ede3_cbc			@33
	hc_EVP_md5				@34
	hc_EVP_rc4				@35
	hc_EVP_sha1				@36
	hc_RAND_status				@37
	hc_RAND_file_name			@38
	hc_RAND_add				@39
	hc_RAND_write_file			@40
	hc_RAND_bytes				@41
	hc_HMAC					@42
	hc_HMAC_CTX_cleanup			@43
	hc_HMAC_CTX_init			@44
	hc_HMAC_Final				@45
	hc_HMAC_Init_ex				@46
	hc_HMAC_Update				@47
	hc_HMAC_size				@48
