EXPORTS
	afsconf_AddKey					@1
	afsconf_AddUser					@2
	afsconf_CellApply				@3
	afsconf_CheckAuth				@4
	afsconf_ClientAuth				@5
	afsconf_ClientAuthSecure			@6
	afsconf_Close					@7
	afsconf_DeleteKey				@8
	afsconf_DeleteUser				@9
	afsconf_GetCellInfo				@10
	afsconf_GetKey					@11
	afsconf_GetKeys					@12
	afsconf_GetLatestKey				@13
	afsconf_GetLocalCell				@14
	afsconf_GetNoAuthFlag				@15
	afsconf_GetNthUser				@16
	afsconf_Open					@18
	afsconf_ServerAuth				@19
	afsconf_SetCellInfo				@20
	afsconf_SetNoAuthFlag				@21
	afsconf_SuperUser				@22
	ka_AuthServerConn				@23
	ka_Authenticate					@24
	ka_CellConfig					@25
	ka_CellToRealm					@26
	ka_ChangePassword				@27
	ka_ExpandCell					@28
	ka_ExplicitCell					@29
	ka_GetAdminToken				@30
	ka_GetAuthToken					@31
	ka_GetSecurity					@32
	ka_GetServerToken				@33
	ka_GetServers					@34
	ka_GetToken					@35
	ka_Init						@36
	ka_LocalCell					@37
	ka_ParseLoginName				@38
	ka_ReadPassword					@39
	ka_SingleServerConn				@40
	ka_StringToKey					@41
	ka_UserAuthenticateGeneral			@42
	ka_VerifyUserToken				@43
	ubik_Call					@44
	ubik_Call_New					@45
	ubik_CallIter					@46
	ubik_ClientDestroy				@47
	ubik_ClientInit					@48
	ubik_ParseClientList				@49
	getDirPath					@50
	ka_AuthSpecificServersConn			@51
;	pthread_recursive_mutex_lock			@52
;	pthread_recursive_mutex_unlock			@53
;	grmutex						@54 DATA
	initAFSDirPath					@55
;     	renamefile					@56
	opendir						@57
	readdir						@58
	closedir					@59
	nterr_nt2unix					@60
        ka_KeyCheckSum                                  @61
	pr_GetCPS					@62
	pr_Initialize					@63
	pr_GetHostCPS					@64
;	afs_krb_get_lrealm				@65
	pr_NameToId					@66
	pr_IdToName					@67
	afs_gettimeofday				@68
	int_to_base32					@69
	base32_to_int					@70
	ktc_ForgetAllTokens				@71
	ktc_ForgetToken					@72
	ktc_GetToken					@73
	ktc_ListTokens					@74
	ktc_SetToken					@75
	ktc_OldPioctl					@76
	pioctl						@77
	rx_Init						@78
	ka_UserAuthenticateGeneral2			@79
    	pr_CreateUser                   		@80
    	pr_SNameToId                    		@81

	DISK_function_names				@83 DATA
	VOTE_function_names				@84 DATA

        pr_End                                          @85
        PR_INewEntry					@86
        PR_WhereIsIt					@87
        PR_DumpEntry					@88
        PR_AddToGroup					@89
        PR_NameToID					@90
        PR_IDToName					@91
        PR_Delete					@92
        PR_RemoveFromGroup				@93
        PR_GetCPS					@94
        PR_NewEntry					@95
        PR_ListMax					@96
        PR_SetMax					@97
        PR_ListEntry					@98
        PR_ChangeEntry					@99
        PR_ListElements					@100
        PR_IsAMemberOf					@101
        PR_SetFieldsEntry				@102
        PR_ListOwned					@103
        PR_GetCPS2					@104
        PR_GetHostCPS					@105
        PR_UpdateEntry					@106
        PR_ListEntries					@107
        PR_ListSuperGroups				@108
;	ka_AuthSpecificServersConn			@109   /* duplicate */
;	ka_KeyCheckSum					@110   /* duplicate */
	rx_Finalize					@111
	rx_InitHost					@112
	VOTE_GetSyncSite				@113
	ubik_RefreshConn				@114
        rx_SetSecurityConfiguration                     @115
        pioctl_utf8                                     @116
        ubik_PR_INewEntry				@117
        ubik_PR_WhereIsIt				@118
        ubik_PR_DumpEntry				@119
        ubik_PR_AddToGroup				@120
        ubik_PR_NameToID				@121
        ubik_PR_IDToName				@122
        ubik_PR_Delete					@123
        ubik_PR_RemoveFromGroup				@124
        ubik_PR_GetCPS					@125
        ubik_PR_NewEntry				@126
        ubik_PR_ListMax					@127
        ubik_PR_SetMax					@128
        ubik_PR_ListEntry				@129
        ubik_PR_ChangeEntry				@130
        ubik_PR_ListElements				@131
        ubik_PR_IsAMemberOf				@132
        ubik_PR_SetFieldsEntry				@133
        ubik_PR_ListOwned				@134
        ubik_PR_GetCPS2					@135
        ubik_PR_GetHostCPS				@136
        ubik_PR_UpdateEntry				@137
        ubik_PR_ListEntries				@138
        ubik_PR_ListSuperGroups				@139
        ugen_ClientInit                                 @140
        afsconf_BuildServerSecurityObjects              @141
        afsconf_ClientAuthToken                         @142
	ktc_GetTokenEx					@143
	ktc_SetTokenEx					@144
	token_extractRxkad				@145
	token_FreeSet					@146
        xdr_ktc_tokenUnion                              @147
        xdr_ktc_setTokenData                            @148
	ktc_ListTokensEx				@149
        pr_SIdToName                                    @150
        afsconf_GetExtendedCellInfo                     @151
        afsconf_UpToDate                                @152
	afsconf_SetSecurityFlags			@153
	afsconf_SetLocalRealm				@154
	afsconf_IsLocalRealmMatch			@155
	afsconf_ParseNetFiles				@156
	afsconf_SetExtendedCellInfo			@157
	afsconf_PickClientSecObj			@158
	afsconf_SawCell					@159    DATA
	ka_PrintBytes					@160
	ka_KeyIsZero					@161
        afsconf_typedKey_new                            @162
        afsconf_typedKey_put                            @163
        afsconf_AddTypedKey                             @164
        afsconf_typedKey_values                         @165
        afsconf_GetAllKeys                              @166
	afsconf_CheckRestrictedQuery			@167
        string_PR_IDToName				@168
        initialize_ACFG_error_table			@169
        initialize_KA_error_table			@170
        initialize_KTC_error_table			@171
        initialize_PT_error_table			@172
        initialize_U_error_table			@173
;        initialize_VL_error_table			@174
;        initialize_VOLS_error_table			@175
        afs_getDirPath					@176
        afsconf_AddIdentity				@177
        afsconf_CellAliasApply				@178
        afsconf_DeleteIdentity				@179
        afsconf_DeleteKeyBySubType			@180
        afsconf_DeleteKeyByType				@181
        afsconf_FindIANAName				@182
        afsconf_FindService				@183
        afsconf_GetAfsdbInfo				@184
        afsconf_GetKeyByTypes				@185
        afsconf_GetKeysByType				@186
        afsconf_GetLatestKeyByTypes			@187
        afsconf_GetLatestKeysByType			@188
        afsconf_GetNthIdentity				@189
        afsconf_IsSuperIdentity				@190
        afsconf_ParseNetRestrictFile			@191
        afsconf_PutTypedKeyList				@192
        afsconf_SuperIdentity				@193
        afsconf_typedKey_free				@194
        afsconf_typedKey_get				@195
        token_SetsEquivalent				@196
        token_addToken					@197
        token_buildTokenJar				@198
        token_findByType				@199
        token_freeToken					@200
        token_freeTokenContents				@201
        token_replaceToken				@202
        token_setPag					@203
        token_setRxkadViceId				@204
        ugen_ClientInitCell				@205
        ugen_ClientInitFlags				@206
        ugen_ClientInitServer				@207
        xdr_idlist					@208
        xdr_namelist					@209
        xdr_prlist					@210
        afsconf_CountKeys				@211
