/*
 * Copyright 2008-2010, Sine Nomine Associates and others.
 * All Rights Reserved.
 *
 * This software has been released under the terms of the IBM Public
 * License.  For details, see the LICENSE file in the top-level source
 * directory or online at http://www.openafs.org/dl/license10.html
 */

#ifndef AFS_UTIL_THREAD_POOL_IMPL_H
#define AFS_UTIL_THREAD_POOL_IMPL_H 1

#include "thread_pool.h"
#include "thread_pool_impl_types.h"

/**
 *  implementation-private interfaces for thread_pool.
 */

#endif /* AFS_UTIL_THREAD_POOL_IMPL_H */
