/*
 * Copyright 2000, International Business Machines Corporation and others.
 * All Rights Reserved.
 *
 * This software has been released under the terms of the IBM Public
 * License.  For details, see the LICENSE file in the top-level source
 * directory or online at http://www.openafs.org/dl/license10.html
 */

#define LOWEST_OPCODE 500
#define PRINEWUSER  500
#define PRWHEREISIT 501
#define PRDUMPENTRY 502
#define PRADDTOGROUP 503
#define PRNAMETOID 504
#define PRIDTONAME 505
#define PRDELETE 506
#define PRREMOVEFROMGROUP 507
#define PRGETCPS 508
#define PRNEWENTRY 509
#define PRLISTMAX 510
#define PRSETMAX 511
#define PRLISTENTRY 512
#define PRCHANGEENTRY 513
#define PRLISTELEMENTS 514
#define PRISAMEMBEROF 515
#define PRSETFIELDSENTRY 516
#define PRLISTOWNED 517
#define PRGETCPS2 518
#define PRGETHOSTCPS 519
#define PRUPDATEENTRY 520
#define PRLISTENTRIES 521
#define PRLISTSUPERGROUPS 530
#define HIGHEST_OPCODE 530
