/*
 * Thanks to Jim Rees and University of Michigan CITI, for the initial
 * OpenBSD porting work.
 */

#ifndef	AFS_AMD64_PARAM_H
#define	AFS_AMD64_PARAM_H

#define SYS_NAME		"amd64_obsd49"
#define SYS_NAME_ID		SYS_NAME_ID_amd64_obsd49

#define AFS_XBSD_ENV		1	/* {Free,Open,Net}BSD */
#define AFS_X86_XBSD_ENV	1
#define AFS_X86_ENV	        1
#define AFS_64BITPOINTER_ENV    1
#define AFS_64BITUSERPOINTER_ENV	1
#define AFSLITTLE_ENDIAN	1

#endif /* AFS_AMD64_PARAM_H */
