#include <sys/fs.h>        /* For UFS_*(fs) macro definitions */

#define BLKSEEK_PROCESSING_ERROR   (-2)
#define BLKSEEK_FILE_WRITEONLY     (-1)
#define BLKSEEK_NOT_ENABLED         0
#define BLKSEEK_ENABLED             1
