EXPORTS
	ktohl					@11
	life_to_time				@12
	rx_DestroyConnection			@13
	rx_EndCall				@14
	rx_Finalize				@15
	rx_GetCachedConnection			@16
	rx_GetCall				@17
	rx_GetIFInfo				@18
	rx_Init					@19
	rx_NewCall				@20
	rx_NewConnection			@21
	rx_NewService				@22
	rx_PrintPeerStats			@23
	rx_PrintStats				@24
	rx_PrintTheseStats			@25
	rx_ReadProc				@26
	rx_ReleaseCachedConnection		@27
	rx_ServerProc				@28
	rx_StartServer				@29
	rx_WriteProc				@30
	rxevent_Init				@31
	rxevent_Post				@32
	rxkad_GetServerInfo			@33
	rxkad_NewClientSecurityObject		@34
	rxkad_NewServerSecurityObject		@35
	rxnull_NewClientSecurityObject		@38
	rxnull_NewServerSecurityObject		@39
	rxs_Release				@40
	time_to_life				@41
	tkt_CheckTimes				@42
	tkt_DecodeTicket			@43
	tkt_MakeTicket				@44
	xdr_array				@45
	xdr_bool				@46
	xdr_bytes				@47
	xdr_char				@48
	xdr_double				@49
	xdr_enum				@50
	xdr_float				@51
	xdr_int					@52
	xdr_long				@53
	xdr_opaque				@54
	xdr_pointer				@55
	xdr_reference				@56
	xdr_short				@57
	xdr_string				@58
	xdr_u_char				@59
	xdr_u_int				@60
	xdr_u_long				@61
	xdr_u_short				@62
	xdr_union				@63
	xdr_vector				@64
	xdr_void				@65
	xdr_wrapstring				@66
	xdrmem_create				@67
	xdrrec_create				@68
	xdrrec_endofrecord			@69
	xdrrec_eof				@70
	xdrrec_skiprecord			@71
	xdrrx_create				@72
	hton_syserr_conv			@73
	rxkad_global_stats			@74 DATA
	afs_com_err				@75
	afs_error_message			@76
	rx_socket				@77 DATA
;	AssertionFailed				@79
	afs_winsockInit				@80
	rxevent_debugFile			@81 DATA
	rx_debugFile				@82 DATA
        rx_connDeadTime                         @83 DATA
	rx_maxReceiveSize			@84 DATA
	rx_UdpBufSize				@85 DATA
	rx_extraQuota				@86 DATA
	rx_extraPackets				@87 DATA
	rx_tranquil				@88 DATA
	rx_getAllAddr				@89
	rx_nWaiting				@90 DATA
	rx_stats				@91 DATA
	rx_SetNoJumbo				@92
	rx_SetConnDeadTime			@93
	rx_FlushWrite				@94
	rx_thread_id_key			@95 DATA
	multi_Finalize				@96
	multi_Select				@97
	multi_Init				@98
	multi_Finalize_Ignore			@99
	afs_add_to_error_table			@100
	xdr_afsUUID				@101
	rx_GetSpecific				@102
	rx_SetSpecific				@103
	rx_KeyCreate				@104
	rx_BusyError				@105 DATA
	rx_BusyThreshold			@106 DATA
	rx_IncrementTimeAndCount		@107
	rx_enable_stats				@108 DATA
	rx_GetServerDebug			@109
	rx_GetServerStats			@110
	rx_GetServerVersion			@111
	rx_GetServerConnections			@112
	rx_stats_mutex				@113 DATA
	rx_GetServerPeers			@114
	rx_RetrieveProcessRPCStats		@115
	rx_RetrievePeerRPCStats			@116
	rx_FreeRPCStats				@117
	rx_queryProcessRPCStats			@118
	rx_queryPeerRPCStats			@119
	rx_enableProcessRPCStats		@120
	rx_enablePeerRPCStats			@121
	rx_disableProcessRPCStats		@122
	rx_disablePeerRPCStats			@123
	RXSTATS_ExecuteRequest			@124
	RXSTATS_RetrieveProcessRPCStats		@125
	RXSTATS_RetrievePeerRPCStats		@126
	RXSTATS_QueryProcessRPCStats		@127
	RXSTATS_QueryPeerRPCStats		@128
	RXSTATS_EnableProcessRPCStats		@129
	RXSTATS_EnablePeerRPCStats		@130
	RXSTATS_DisableProcessRPCStats		@131
	RXSTATS_DisablePeerRPCStats		@132
	RXSTATS_QueryRPCStatsVersion		@133
	RXSTATS_ClearProcessRPCStats		@134
	RXSTATS_ClearPeerRPCStats		@135
	rx_clearProcessRPCStats                 @136
	rx_clearPeerRPCStats                 	@137
	rxi_connAbortThreshhold			@138 DATA
	rxi_connAbortDelay			@139 DATA
	rxi_callAbortThreshhold			@140 DATA
	rxi_callAbortDelay			@141 DATA
	RXSTATS_function_names			@142 DATA
;	RXAFSCB_function_names			@143 DATA
;	RXAFSCB_CallBack			@144
;	StartRXAFSCB_CallBack			@145
;	EndRXAFSCB_CallBack			@146
;	RXAFSCB_InitCallBackState		@147
;	RXAFSCB_Probe				@148
;	StartRXAFSCB_Probe			@149
;	EndRXAFSCB_Probe			@150
;	RXAFSCB_GetLock				@151
;	RXAFSCB_GetCE				@152
;	RXAFSCB_XStatsVersion			@153
;	RXAFSCB_GetXStats			@154
;	RXAFSCB_InitCallBackState2		@155
;	RXAFSCB_WhoAreYou			@156
;	RXAFSCB_InitCallBackState3		@157
;	RXAFSCB_ProbeUuid			@158
;	StartRXAFSCB_ProbeUuid			@159
;	EndRXAFSCB_ProbeUuid			@160
;	RXAFSCB_GetServerPrefs			@161
;	RXAFSCB_GetCellServDB			@162
;	RXAFSCB_GetLocalCell			@163
;	RXAFSCB_GetCacheConfig			@164
;	StartRXAFS_FetchData			@165
;	EndRXAFS_FetchData			@166
;	RXAFS_FetchACL				@167
;	RXAFS_FetchStatus			@168
;	StartRXAFS_StoreData			@169
;	EndRXAFS_StoreData			@170
;	RXAFS_StoreACL				@171
;	RXAFS_StoreStatus			@172
;	RXAFS_RemoveFile			@173
;	RXAFS_CreateFile			@174
;	RXAFS_Rename				@175
;	RXAFS_Symlink				@176
;	RXAFS_Link				@177
;	RXAFS_MakeDir				@178
;	RXAFS_RemoveDir				@179
;	RXAFS_OldSetLock			@180
; 	RXAFS_OldExtendLock			@181
;	RXAFS_OldReleaseLock			@182
;	RXAFS_GetStatistics			@183
;	RXAFS_GiveUpCallBacks			@184
;	RXAFS_GetVolumeInfo			@185
;	RXAFS_GetVolumeStatus			@186
;	RXAFS_SetVolumeStatus			@187
;	RXAFS_GetRootVolume			@188
;	RXAFS_CheckToken			@189
;	RXAFS_GetTime				@190
;	RXAFS_NGetVolumeInfo			@191
;	RXAFS_BulkStatus			@192
;	RXAFS_SetLock				@193
;	RXAFS_ExtendLock			@194
;	RXAFS_ReleaseLock			@195
;	RXAFS_XStatsVersion			@196
;	RXAFS_GetXStats				@197
;	RXAFS_Lookup				@198
;	RXAFS_FlushCPS				@199
;	RXAFS_DFSSymlink			@200
	rx_SetRxStatUserOk			@201
	rx_RxStatUserOk				@202
	rx_enable_hot_thread			@203 DATA
	xdr_int64				@204
	xdr_uint64				@205
	rx_SetMaxMTU				@206
	rx_GetConnection			@207
;       RXAFSCB_GetCE64                         @208
;       RXAFSCB_GetCellByNum                    @209
;       RXAFSCB_TellMeAboutYourself             @210
;       EndRXAFS_FetchData64                    @211
;       EndRXAFS_StoreData64                    @212
;       RXAFS_CallBackRxConnAddr                @213
;       RXAFS_GetCapabilities                   @214
;       RXAFS_GiveUpAllCallBacks                @215
;       RXAFS_InlineBulkStatus                  @216
;       RXAFS_ResidencyCmd                      @217
;       StartRXAFS_FetchData64                  @218
;       StartRXAFS_StoreData64                  @219
        rx_StartClientThread                    @220
	rxkad_global_stats_lock			@221 DATA
	rxkad_stats_key				@222 DATA
	rx_DebugOnOff				@223
	rx_InitHost				@224
	rx_NewServiceHost			@225
        osi_AssertFailU				@226
	DllMain					@227
        rx_SetSecurityConfiguration             @228
;       RXAFSCB_ExecuteRequest                  @229
;       EndRXAFS_GetTime                        @230
;       StartRXAFS_GetTime                      @231
;       EndRXAFS_GetCapabilities                @232
;       StartRXAFS_GetCapabilities              @233
        xdr_afs_uint32                          @234
        xdr_afs_int32                           @235
        xdr_afs_uint64                          @236
        xdr_afs_int64                           @237
        clock_now                               @238
        clock_UpdateTime                        @239
        clock_haveCurrentTime                   @240
        rx_SetRxDeadTime                        @241
        rx_GetMinUdpBufSize                     @242
        rx_SetUdpBufSize                        @243        
        rx_getAllAddrMaskMtu                    @244        
        rx_stats_active                         @245 DATA
        rx_StatsOnOff                           @246
        rx_ReadvProc                            @247
        rx_WritevProc                           @248
        rx_WritevAlloc                          @249
        rx_ReadProc32                           @250
        rx_WriteProc32                          @251
        xdr_free                                @252
;       RXAFS_FsCmd                             @253
        rxi_CallError                           @254
        rx_GetLocalPeers			@255
        xdr_alloc                               @256
	rx_GetMaxReceiveWindow			@257
	rx_GetMaxSendWindow			@258
	rx_SetMaxReceiveWindow			@259
	rx_SetMaxSendWindow			@260
	rx_GetMinPeerTimeout			@261
	rx_SetMinPeerTimeout			@262
;       xdr_serverList                          @263
;       xdr_Capabilities                        @264
        xdrlen_create                           @265
        rx_SetConnSecondsUntilNatPing           @266
        rx_GetServiceSpecific                   @267
        rx_SetServiceSpecific                   @268
	rx_NewThreadId				@269
        rx_GetStatistics                        @270
        rx_FreeStatistics                       @271
	rx_SetConnHardDeadTime			@272
	rx_SetConnIdleDeadTime			@273
	rx_InterruptCall                        @274
        osi_Panic				@275

; lwp functions
        Afs_Lock_Obtain                         @276
        Afs_Lock_ReleaseR                       @277
        Afs_Lock_ReleaseW                       @278
        Afs_Lock_WakeupR                        @279
        Lock_Init                               @280
        Lock_Destroy                            @281
        FT_Init                                 @282
        FT_GetTimeOfDay                         @283
        FT_AGetTimeOfDay                        @284
        FT_ApproxTime                           @285
        threadname                              @286
        registerthread                          @287
        swapthreadname                          @288

	rx_opaque_new                           @289
	rx_opaque_alloc                         @290
	rx_opaque_populate                      @291
	rx_opaque_copy                          @292
	rx_opaque_freeContents                  @293
	rx_opaque_zeroFreeContents              @294
	rx_opaque_free                          @295
	rx_opaque_zeroFree                      @296
	rx_identity_match                       @297
	rx_identity_populate                    @298
	rx_identity_copy                        @299
	rx_identity_copyContents                @300
	rx_identity_new                         @301
	rx_identity_freeContents                @302
	rx_identity_free                        @303

;	rx_SetBusyChannelError                  @304
	rx_GetThreadNum				@305
	rx_SetThreadNum				@306

;	rx_SetConnectionEpoch			@307
        rx_GetConnectionEpoch                   @308
;	rx_SetConnectionId			@309
        rx_GetConnectionId                      @310
        rx_GetSecurityData                      @311
        rx_SetSecurityData                      @312
        rx_IsUsingPktCksum                      @313
        rx_SetSecurityHeaderSize                @314
        rx_GetSecurityHeaderSize                @315
        rx_SetSecurityMaxTrailerSize            @316
        rx_GetSecurityMaxTrailerSize            @317
        ; @318 is available
        rx_SetMsgsizeRetryErr                   @319
        rx_IsServerConn                         @320
        rx_IsClientConn                         @321
        rx_SecurityObjectOf                     @322
        rx_PeerOf                               @323
        rx_ServiceIdOf                          @324
        rx_SecurityClassOf                      @325
        rx_ServiceOf                            @326
        rx_ConnError                            @327

        rx_ConnectionOf                         @328
        rx_Error                                @329
        rx_GetRemoteStatus                      @330
        rx_SetLocalStatus                       @331
        rx_GetCallAbortCode                     @332
        rx_SetCallAbortCode                     @333
        rx_RecordCallStatistics                 @334
;	rx_KeepAliveOn				@335
;	rx_KeepAliveOff				@336
	rx_HostOf				@337
        rx_PortOf                               @338

        rx_ClearProcessRPCStats                 @339
        rx_ClearPeerRPCStats                    @340
        rx_CopyProcessRPCStats                  @341
        rx_CopyPeerRPCStats                     @342
	rx_ReleaseRPCStats                      @343
;       RXAFS_OpCodeIndex                       @344
	initialize_RXK_error_table		@345
	rx_GetNetworkError                      @346
        afs_set_com_err_hook                    @347
	rxkad_NewKrb5ServerSecurityObject       @348
	tkt_MakeTicket5                         @349
	tkt_DeriveDesKey                        @350
;	opr_AssertFailU                         @351
;	opr_AssertionFailed                     @352
;	xdr_Capabilities                        @353
	xdr_rpcStats                            @354
	rx_GetCallStatus                        @355

; for performance testing
        rx_TSFPQGlobSize                        @2001 DATA
        rx_TSFPQLocalMax                        @2002 DATA
        rx_TSFPQMaxProcs                        @2003 DATA
        rxi_MorePackets                         @2007
        rxi_nRecvFrags                          @2008 DATA
        rxi_nSendFrags                          @2009 DATA
        rx_initReceiveWindow                    @2010 DATA
        rx_initSendWindow                       @2011 DATA
        rx_intentionallyDroppedPacketsPer100    @2012 DATA
        rx_intentionallyDroppedOnReadPer100     @2013 DATA

; for debugging
        rx_DumpCalls                            @9998
        rx_DumpPackets                          @9999
