/*
 *     Generated by class-dump 3.1.1.
 *
 *     class-dump is Copyright (C) 1997-1998, 2000-2001, 2004-2006 by Steve Nygard.
 */
#import <AppKit/AppKit.h>

@interface NSMenuExtra : NSStatusItem
{
    NSBundle *_bundle;
    NSMenu *_menu;
    NSView *_view;
    float _length;
    struct {
        unsigned int customView:1;
        unsigned int menuDown:1;
        unsigned int reserved:30;
    } _flags;
    id _controller;
}

- (id)initWithBundle:(id)fp8;
- (id)initWithBundle:(id)fp8 data:(id)fp12;
- (void)willUnload;
- (void)dealloc;
- (id)bundle;
- (float)length;
- (void)setLength:(float)fp8;
- (id)image;
- (void)setImage:(id)fp8;
- (id)alternateImage;
- (void)setAlternateImage:(id)fp8;
- (id)menu;
- (void)setMenu:(id)fp8;
- (id)toolTip;
- (void)setToolTip:(id)fp8;
- (id)view;
- (void)setView:(id)fp8;
- (BOOL)isMenuDown;
- (void)drawMenuBackground:(BOOL)fp8;
- (void)popUpMenu:(id)fp8;
- (void)unload;
- (id)statusBar;
- (SEL)action;
- (void)setAction:(SEL)fp8;
- (id)target;
- (void)setTarget:(id)fp8;
- (id)title;
- (void)setTitle:(id)fp8;
- (id)attributedTitle;
- (void)setAttributedTitle:(id)fp8;
- (BOOL)isEnabled;
- (void)setEnabled:(BOOL)fp8;
- (void)setHighlightMode:(BOOL)fp8;
- (BOOL)highlightMode;
- (void)sendActionOn:(int)fp8;
- (id)_initInStatusBar:(id)fp8 withLength:(float)fp12 withPriority:(int)fp16;
- (id)_window;
- (id)_button;
- (void)_adjustLength;

@end

@interface NSMenuExtra (NSMenuExtraPrivate)
+ (unsigned int)defaultLength;
- (void)setController:(id)fp8;
- (id)controller;
- (void)setMenuDown:(BOOL)fp8;
- (float)defaultLength;
- (id)accessibilityAttributeNames;
- (id)accessibilityAttributeValue:(id)fp8;
- (BOOL)accessibilityIsAttributeSettable:(id)fp8;
- (void)accessibilitySetValue:(id)fp8 forAttribute:(id)fp12;
- (id)accessibilityActionNames;
- (id)accessibilityActionDescription:(id)fp8;
- (void)accessibilityPerformAction:(id)fp8;
- (BOOL)accessibilityIsIgnored;
- (id)accessibilityHitTest:(struct _NSPoint)fp8;
- (id)accessibilityFocusedUIElement;
- (id)AXRole;
- (id)AXRoleDescription;
- (id)AXSubrole;
- (id)AXDescription;
- (id)AXChildren;
- (id)AXParent;
- (id)AXTitle;
- (id)AXValue;
- (id)AXEnabled;
- (id)AXSelected;
- (BOOL)isAXSelectedSettable;
- (void)setAXSelected:(id)fp8;
- (id)AXPosition;
- (id)AXSize;
- (void)performAXPress;
- (void)performAXCancel;
@end

@interface NSDockExtra : NSObject
{
    NSBundle *_bundle;
    unsigned int _id;
    void *_nativeWindow;
    id _controller;
    struct {
        unsigned int nativeWindow:1;
        unsigned int reserved:31;
    } _flags;
    long _fReserved1;
    long _fReserved2;
    long _fReserved3;
    long _fReserved4;
}

- (id)initWithBundle:(id)fp8;
- (id)initWithBundle:(id)fp8 data:(id)fp12;
- (void)dealloc;
- (void)willUnload;
- (id)bundle;
- (id)menu;
- (void)toggle;
- (void)handleMenuCommand:(unsigned int)fp8 tag:(unsigned int)fp12;
- (void)setDockLabel:(id)fp8;
- (id)cocoaWindow;
- (void *)carbonCGrafPtr;

@end

@interface NSDockExtra (NSDockExtraReallyPrivate)
- (void)_releaseWindow;
@end

@interface NSDockExtra (NSDockExtraPrivate)
- (id)initWithBundle:(id)fp8 identifier:(unsigned int)fp12 controller:(id)fp16;
- (id)initWithBundle:(id)fp8 identifier:(unsigned int)fp12 data:(id)fp16 controller:(id)fp20;
- (void *)cgWindow;
@end

@interface NSMutableArray (NSDockExtraMenu)
- (void)insertItem:(id)fp8 atIndex:(int)fp12;
- (void)addItem:(id)fp8;
- (id)insertItemWithTitle:(id)fp8 action:(SEL)fp12 atIndex:(int)fp16;
- (id)addItemWithTitle:(id)fp8 action:(SEL)fp12;
- (void)removeItemAtIndex:(int)fp8;
- (void)setSubmenu:(id)fp8 forItem:(id)fp12;
- (int)indexOfItemWithTitle:(id)fp8;
- (int)indexOfItemWithAction:(SEL)fp8;
- (id)itemWithTitle:(id)fp8;
- (id)itemWithAction:(SEL)fp8;
@end

@interface NSMutableDictionary (NSDockExtraMenuItem)
+ (id)separatorItem;
+ (id)itemWithTitle:(id)fp8 action:(SEL)fp12;
- (BOOL)hasSubmenu;
- (void)setSubmenu:(id)fp8;
- (id)submenu;
- (void)setSubmenuSelectable:(BOOL)fp8;
- (BOOL)isSubmenuSelectable;
- (void)setTitle:(id)fp8;
- (id)title;
- (BOOL)isSeparatorItem;
- (void)setEnabled:(BOOL)fp8;
- (BOOL)isEnabled;
- (void)setAction:(SEL)fp8;
- (SEL)action;
- (void)setTag:(int)fp8;
- (int)tag;
- (void)setMark:(int)fp8;
- (int)mark;
- (void)setMarkChar:(id)fp8;
- (id)markChar;
- (void)setHeader:(BOOL)fp8;
- (BOOL)isHeader;
- (void)setStyle:(int)fp8;
- (int)style;
- (void)setIndent:(unsigned short)fp8;
- (unsigned short)indent;
- (void)setDynamicMask:(int)fp8;
- (int)dynamicMask;
- (void)removeDynamicMask;
- (void)setDynamicBreak:(BOOL)fp8;
- (BOOL)dynamicBreak;
- (void)setSystemIcon:(unsigned int)fp8;
- (unsigned int)systemIcon;
- (void)setResourceIcon:(id)fp8;
- (id)resourceIcon;
- (void)setIconEnabled:(BOOL)fp8;
- (BOOL)isIconEnabled;
@end

@interface NSApplicationDockExtra : NSDockExtra
{
}

- (void)applicationLaunched;
- (void)applicationDied;
- (void)setDockLabel:(id)fp8;

@end

@interface NSMenuExtraView : NSView
{
    NSMenu *_menu;
    NSMenuExtra *_menuExtra;
    NSImage *_image;
    NSImage *_alternateImage;
}

- (id)initWithFrame:(NSRect)fp8 menuExtra:(id)fp24;
- (void)dealloc;
- (void)setMenu:(id)fp8;
- (id)image;
- (void)setImage:(id)fp8;
- (id)alternateImage;
- (void)setAlternateImage:(id)fp8;
- (void)drawRect:(NSRect)fp8;
- (void)mouseDown:(id)fp8;

@end

